#include "elf.h"
#include "dwarf.c"

void readElf(char *fnam){
	int c;
	FILE *f;
	char st[PATH_MAX],*shstrtab,*cp;
	struct elf32_hdr ehdr;
	Elf32_Shdr *shdr;
	stab_rec *stab;
	int stabsize,r;
	char *stabstr;
	symtab_rec *symtab;
	int		symtabsize;
	static char	*strtab;

	stab=0;
	stabstr=0;
	symtab=0;
	f=fopen(fnam,"rb");
	if(!f) return;
	//printf("%s\n",fnam);
	r=fread(&ehdr,sizeof(Elf32_Ehdr),1,f);
	if(ehdr.e_machine!=83){
		printf("illegal machine=%d\n",ehdr.e_machine);
		return;
	}

	if(strtab){
		free(strtab);
		strtab=0;
	}
	shdr=(Elf32_Shdr*) calloc(ehdr.e_shnum,sizeof(Elf32_Shdr));
	fseek(f,ehdr.e_shoff,SEEK_SET);
	r=fread(shdr,sizeof(Elf32_Shdr),ehdr.e_shnum,f);
	shstrtab=(char*) malloc(shdr[ehdr.e_shstrndx].sh_size);
	fseek(f,shdr[ehdr.e_shstrndx].sh_offset,SEEK_SET);
	r=fread(shstrtab,shdr[ehdr.e_shstrndx].sh_size,1,f);

	for(c=0;c<ehdr.e_shnum;c++){
		strcpy(st,shstrtab+shdr[c].sh_name);
		//printf("Shdr%2d: name %04x typ %04x offset %04x addr %06x size %04x %s\n",c,shdr[c].sh_name,shdr[c].sh_type,shdr[c].sh_offset,shdr[c].sh_addr,shdr[c].sh_size,st);
		if(!strcmp(st,".text")){
			unsigned char *cp;
			unsigned int size,ctr,dp;
			size=shdr[c].sh_size;
			cp=(unsigned char*) malloc(size);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(cp,size,1,f);
			dp=shdr[c].sh_addr;
			//printf("Text size %x %d\n",size,size);
			for(ctr=0;ctr<size;ctr+=2){
				flash[dp].opcode=((cp[ctr])<<0)+((cp[ctr+1])<<8);
				//printf("%04x %04x\n",dp,flash[dp].opcode);
				dp++;
			}
			free(cp);
		}

		if(!strcmp(st,".debug_line")){
			unsigned char *cp;
			unsigned int size;
			size=shdr[c].sh_size;
			cp=(unsigned char*) malloc(size);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(cp,size,1,f);
			decode_debug_line(cp,size,fnam);
			free(cp);
		}

		if(!strcmp(st,".stab")){
			stab=(stab_rec*) malloc(shdr[c].sh_size);
			stabsize=shdr[c].sh_size/sizeof(stab_rec);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(stab,shdr[c].sh_size,1,f);
		}

		if(!strcmp(st,".symtab")){
			symtab=(symtab_rec*) malloc(shdr[c].sh_size);
			symtabsize=shdr[c].sh_size/sizeof(symtab_rec);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(symtab,shdr[c].sh_size,1,f);
		}

		if(!strcmp(st,".stabstr")){
			stabstr=(char*) malloc(shdr[c].sh_size);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(stabstr,shdr[c].sh_size,1,f);
		}

		if(!strcmp(st,".strtab")){
			strtab=(char*) malloc(shdr[c].sh_size);
			fseek(f,shdr[c].sh_offset,SEEK_SET);
			r=fread(strtab,shdr[c].sh_size,1,f);
		}

	}
	fclose(f);

	if(symtab){
		for(c=0;c<symtabsize;c++){
			if(symtab[c].name){
				switch(symtab[c].flag2){
					case 0x01:
					//printf("Code %04x: %s\n",symtab[c].value,strtab+symtab[c].name);
					flash[(symtab[c].value)>>1].name=strtab+symtab[c].name;
					break;

					default:
					//printf("%3d: value=%08x size=%3d type=%02x %02x %02x %02x name=%s\n",c,symtab[c].value,symtab[c].size,symtab[c].typ,symtab[c].flag1,symtab[c].flag2,symtab[c].flag3,strtab+symtab[c].name);
					break;

				}
			}
		}
	}

	if(stab){
		int adr=0,radr,fileno;
		char *cp;
		for(c=0;c<stabsize;c++){
			{
				switch(stab[c].typ){

					case 0x84:
					strcpy(st,stabstr+stab[c].name);
					completeFilename(st,fnam,includePath);
					fileno=filenNameFuel();
					addFileName(st);
					adr=-1;
					//printf("neues Sourcefile %d: %s\n",fileno,st);
					break;

					case 0x24:
					adr=stab[c].adr;
					//printf("neue Funktion: %s %04x\n",stabstr+stab[c].name,adr);
					break;

					case 0x44:
					flash[(adr+stab[c].adr)>>1].src_file=fileno;
					flash[(adr+stab[c].adr)>>1].src_line=stab[c].line;
					flash[(adr+stab[c].adr)>>1].flags|=PM_SOURCE;
					if(adr<0) radr=(stab[c].adr)>>1;
					else radr=(stab[c].adr+adr)>>1;
					cp=strrchr(st,'/');
					if(cp) cp++;
					else cp=st;
					//printf("SrcLine Adr:%04x (%04x) %5d %s\n",radr,radr<<1,stab[c].line,cp);
					break;

					default:
					//printf("%3d:adr=%08x typ=%02X line=%4d name=%s\n",c,stab[c].adr,stab[c].typ,stab[c].line,stabstr+stab[c].name);
					break;

				}
			}
		}
	}

	if(stab) free(stab);
	if(stabstr) free(stabstr);
	if(symtab) free(symtab);

	strcpy(fn_hexFile,fnam);
	strcpy(st,fnam);
	cp=strrchr(st,'/');
	if(!cp) cp=strrchr(st,'\\');
	if(!cp) cp=st; else cp++;
	strcpy(fileshortname,cp);
	refreshTitle();
}
